Package: DataDNA
Title: Data Frame Fingerprints and Lineage Figures
Version: 0.1.0
Authors@R: c(
    person("Tony", "Lu", role = c("aut", "cre"),
           email = "xulunt123@gmail.com")
    )
Description: Profiles R data frames as compact data fingerprints using schema,
    shape, missingness, distribution, category, uniqueness, time, and role
    signals. It compares versions, identifies close relatives in a library of
    historical data sets, and renders portable HTML cards plus static PNG/PDF
    lineage figures for reports.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.2
Depends: R (>= 4.1.0)
Imports: htmltools
NeedsCompilation: no
Packaged: 2026-05-04 10:06:39 UTC; TonyL
Author: Tony Lu [aut, cre]
Maintainer: Tony Lu <xulunt123@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-06 20:00:02 UTC
Built: R 4.5.2; ; 2026-05-06 20:34:43 UTC; unix

export(data_dna)
export(dna_card)
export(dna_compare)
export(dna_diff)
export(dna_match)
export(dna_match_card)
export(dna_match_plot)
export(dna_species)
export(dna_example_customers)
S3method(print,data_dna)
S3method(print,dna_comparison)
S3method(print,dna_diff)
S3method(print,dna_match)
importFrom(utils,head)

# DataDNA

**DataDNA** is an R package that gives every data frame a compact fingerprint,
lineage match, and report-ready identity figure.

Instead of only asking "what is in this table?", DataDNA asks:

- What kind of data set is this?
- How stable is its identity?
- Did this version drift from the previous one?
- Which columns changed their role, missingness, categories, or distribution?

The package is designed for analysts who receive CSVs, extracts, dashboards, or
modeling data sets and need a fast way to recognize and compare them.

## Example

```r
library(DataDNA)

demo <- dna_example_customers()

dna <- data_dna(demo$customers_new, name = "customers_new")
dna

card <- dna_card(dna, file = "customers_dna.html")

dna_compare(demo$customers_old, demo$customers_new)
dna_diff(demo$customers_old, demo$customers_new)
```

`dna_compare()` combines exact schema overlap with shape, species, role
structure, distribution, missingness, category, and identity signals. This makes
the score feel more like a data fingerprint than a strict column-name check.

The package also includes lazy-loaded `customers_old` and `customers_new`
example data sets.

## Find the closest ancestor

```r
library <- list(
  customers_2024 = data_dna(customers_old),
  customers_2025 = data_dna(customers_new)
)

match <- dna_match(customers_new, library)
match

dna_match_plot(match, file = "lineage.png")
```

`dna_match_plot()` is now the recommended reporting output. It renders a static
PNG/PDF lineage figure with base R graphics: white background, compact ranking
table, and restrained similarity lines that fit technical reports, papers, and
slide decks better than a web page.

## Core API

```r
data_dna(df)
dna_card(df)
dna_compare(old_df, new_df)
dna_diff(old_df, new_df)
dna_match(new_df, dna_library)
dna_match_card(match)
dna_match_plot(match)
dna_species(df)
```

## Installation

From GitHub:

```r
install.packages("devtools")
devtools::install_github("TonyIsFool/DataDNA")
```

Or with the lighter `remotes` package:

```r
install.packages("remotes")
remotes::install_github("TonyIsFool/DataDNA")
```

From a local source tarball:

```r
install.packages("DataDNA_0.1.0.tar.gz", repos = NULL, type = "source")
```

## Design

The profiling and comparison algorithms use base R. The HTML card uses the
lightweight `htmltools` package so the result is portable and CRAN-friendly.

#' Example customer tables
#'
#' Creates two small customer data frames designed to demonstrate DataDNA cards,
#' comparison, and mutation reports.
#'
#' @return A list with `customers_old` and `customers_new` data frames.
#' @export
#'
#' @examples
#' demo <- dna_example_customers()
#' str(demo$customers_old)
dna_example_customers <- function() {
  set.seed(42)
  n <- 180
  customers_old <- data.frame(
    customer_id = sprintf("C%04d", seq_len(n)),
    signup_date = as.Date("2025-01-01") + sample(0:240, n, replace = TRUE),
    region = sample(c("North", "South", "East", "West"), n, replace = TRUE, prob = c(0.28, 0.26, 0.24, 0.22)),
    segment = sample(c("Starter", "Growth", "Enterprise"), n, replace = TRUE, prob = c(0.62, 0.28, 0.10)),
    age = round(stats::rnorm(n, 38, 11)),
    income = round(stats::rlnorm(n, log(62000), 0.35)),
    active = sample(c(TRUE, FALSE), n, replace = TRUE, prob = c(0.76, 0.24)),
    churn_target = sample(c("no", "yes"), n, replace = TRUE, prob = c(0.83, 0.17)),
    stringsAsFactors = FALSE
  )

  customers_new <- customers_old
  customers_new$income <- round(customers_new$income * stats::runif(n, 1.02, 1.22))
  customers_new$region[sample(seq_len(n), 18)] <- "International"
  customers_new$age[sample(seq_len(n), 28)] <- NA
  customers_new$loyalty_score <- round(stats::runif(n, 0, 100), 1)
  customers_new$active <- sample(c(TRUE, FALSE), n, replace = TRUE, prob = c(0.69, 0.31))

  list(customers_old = customers_old, customers_new = customers_new)
}

#' Create a data DNA profile
#'
#' Profiles an R data frame into a compact identity object that records schema,
#' shape, missingness, distributions, categories, uniqueness, time signals, and
#' stable fingerprints.
#'
#' @param df A data frame.
#' @param name Optional data set name shown on cards and print output.
#' @param sample_size Maximum number of rows used for profiling.
#'
#' @return A `data_dna` object.
#' @export
#'
#' @examples
#' demo <- dna_example_customers()
#' dna <- data_dna(demo$customers_new, name = "customers_new")
#' dna
data_dna <- function(df, name = NULL, sample_size = 10000L) {
  if (!is.data.frame(df)) {
    stop("`df` must be a data frame.", call. = FALSE)
  }
  stopifnot(length(sample_size) == 1L, sample_size >= 1L)

  n_rows <- nrow(df)
  n_cols <- ncol(df)
  idx <- sample_index(n_rows, sample_size)
  sample_df <- df[idx, , drop = FALSE]
  profiles <- lapply(names(sample_df), function(nm) column_profile(sample_df[[nm]], nm, n_rows = n_rows))
  columns <- profiles_to_frame(profiles)
  species <- dna_species_from_profile(columns, n_rows, n_cols)
  scores <- dna_scores(columns, n_rows, n_cols)
  fingerprints <- make_fingerprints(columns, n_rows, n_cols)
  dna_id <- compact_id(fingerprints$schema_hash, fingerprints$shape_hash, fingerprints$profile_hash)
  signals <- dna_signals(columns, n_rows, n_cols, species)
  warnings <- dna_warnings(columns)

  structure(
    list(
      name = if (is.null(name)) deparse(substitute(df)) else as.character(name),
      dna_id = dna_id,
      species = species,
      n_rows = n_rows,
      n_cols = n_cols,
      sample_size = length(idx),
      columns = columns,
      scores = scores,
      fingerprints = fingerprints,
      signals = signals,
      warnings = warnings,
      created = Sys.time()
    ),
    class = "data_dna"
  )
}

#' Guess the species of a data frame
#'
#' @param df A data frame.
#'
#' @return A character label such as `customer_table`, `event_stream`, or
#'   `wide_feature_matrix`.
#' @export
#'
#' @examples
#' dna_species(dna_example_customers()$customers_new)
dna_species <- function(df) {
  if (inherits(df, "data_dna")) {
    return(df$species)
  }
  if (!is.data.frame(df)) {
    stop("`df` must be a data frame or data_dna object.", call. = FALSE)
  }
  profiles <- lapply(names(df), function(nm) column_profile(df[[nm]], nm, n_rows = nrow(df)))
  dna_species_from_profile(profiles_to_frame(profiles), nrow(df), ncol(df))
}

#' @export
print.data_dna <- function(x, ...) {
  cat("DataDNA Card\n")
  cat("  ID: ", x$dna_id, "\n", sep = "")
  cat("  Name: ", x$name, "\n", sep = "")
  cat("  Species: ", x$species, "\n", sep = "")
  cat("  Shape: ", format(x$n_rows, big.mark = ","), " rows x ", x$n_cols, " columns\n", sep = "")
  cat("  Identity strength: ", round(x$scores$identity_strength), " / 100\n", sep = "")
  cat("  Stability: ", x$scores$stability, "\n", sep = "")
  cat("  Drift risk: ", x$scores$drift_risk, "\n", sep = "")
  if (length(x$signals)) {
    cat("  Signals: ", paste(head(x$signals, 3L), collapse = " | "), "\n", sep = "")
  }
  if (length(x$warnings)) {
    cat("  Warnings: ", paste(head(x$warnings, 3L), collapse = " | "), "\n", sep = "")
  }
  invisible(x)
}

sample_index <- function(n_rows, sample_size) {
  if (n_rows == 0L) {
    return(integer())
  }
  size <- min(n_rows, as.integer(sample_size))
  unique(as.integer(round(seq(1, n_rows, length.out = size))))
}

column_profile <- function(x, name, n_rows) {
  type <- column_type(x)
  missing_rate <- mean(is.na(x))
  non_missing <- x[!is.na(x)]
  n_unique <- length(unique(non_missing))
  unique_rate <- if (length(non_missing)) n_unique / length(non_missing) else NA_real_
  role <- infer_role(name, type, missing_rate, unique_rate, n_unique, n_rows, non_missing)

  base <- list(
    name = name,
    type = type,
    role = role,
    missing_rate = missing_rate,
    unique_rate = unique_rate,
    n_unique = n_unique,
    mean = NA_real_,
    sd = NA_real_,
    q0 = NA_real_,
    q25 = NA_real_,
    q50 = NA_real_,
    q75 = NA_real_,
    q100 = NA_real_,
    zero_rate = NA_real_,
    outlier_rate = NA_real_,
    entropy = NA_real_,
    tail_rate = NA_real_,
    top_values = list(character()),
    date_min = NA_character_,
    date_max = NA_character_,
    interval_cv = NA_real_
  )

  if (type == "numeric") {
    vals <- as.numeric(non_missing)
    qs <- safe_quantile(vals)
    iqr <- qs[[4L]] - qs[[2L]]
    lower <- qs[[2L]] - 1.5 * iqr
    upper <- qs[[4L]] + 1.5 * iqr
    base$mean <- safe_mean(vals)
    base$sd <- safe_sd(vals)
    base$q0 <- qs[[1L]]
    base$q25 <- qs[[2L]]
    base$q50 <- qs[[3L]]
    base$q75 <- qs[[4L]]
    base$q100 <- qs[[5L]]
    base$zero_rate <- if (length(vals)) mean(vals == 0) else NA_real_
    base$outlier_rate <- if (length(vals) && iqr > 0) mean(vals < lower | vals > upper) else 0
  } else if (type %in% c("character", "factor", "logical")) {
    vals <- as.character(non_missing)
    base$entropy <- category_entropy(vals)
    base$tail_rate <- category_tail_rate(vals)
    base$top_values <- list(top_values(vals))
  } else if (type == "date") {
    vals <- sort(as.numeric(as.POSIXct(non_missing)))
    base$date_min <- if (length(vals)) as.character(min(non_missing)) else NA_character_
    base$date_max <- if (length(vals)) as.character(max(non_missing)) else NA_character_
    diffs <- diff(vals)
    base$interval_cv <- if (length(diffs) > 1L && mean(diffs) != 0) stats::sd(diffs) / abs(mean(diffs)) else 0
  }

  base
}

profiles_to_frame <- function(profiles) {
  if (!length(profiles)) {
    return(data.frame())
  }
  data.frame(
    name = vapply(profiles, `[[`, character(1), "name"),
    type = vapply(profiles, `[[`, character(1), "type"),
    role = vapply(profiles, `[[`, character(1), "role"),
    missing_rate = vapply(profiles, `[[`, numeric(1), "missing_rate"),
    unique_rate = vapply(profiles, `[[`, numeric(1), "unique_rate"),
    n_unique = vapply(profiles, `[[`, integer(1), "n_unique"),
    mean = vapply(profiles, `[[`, numeric(1), "mean"),
    sd = vapply(profiles, `[[`, numeric(1), "sd"),
    q0 = vapply(profiles, `[[`, numeric(1), "q0"),
    q25 = vapply(profiles, `[[`, numeric(1), "q25"),
    q50 = vapply(profiles, `[[`, numeric(1), "q50"),
    q75 = vapply(profiles, `[[`, numeric(1), "q75"),
    q100 = vapply(profiles, `[[`, numeric(1), "q100"),
    zero_rate = vapply(profiles, `[[`, numeric(1), "zero_rate"),
    outlier_rate = vapply(profiles, `[[`, numeric(1), "outlier_rate"),
    entropy = vapply(profiles, `[[`, numeric(1), "entropy"),
    tail_rate = vapply(profiles, `[[`, numeric(1), "tail_rate"),
    top_values = I(lapply(profiles, `[[`, "top_values")),
    date_min = vapply(profiles, `[[`, character(1), "date_min"),
    date_max = vapply(profiles, `[[`, character(1), "date_max"),
    interval_cv = vapply(profiles, `[[`, numeric(1), "interval_cv"),
    stringsAsFactors = FALSE
  )
}

column_type <- function(x) {
  if (inherits(x, c("Date", "POSIXct", "POSIXlt"))) {
    "date"
  } else if (is.logical(x)) {
    "logical"
  } else if (is.factor(x)) {
    "factor"
  } else if (is.numeric(x) || is.integer(x)) {
    "numeric"
  } else if (is.character(x)) {
    "character"
  } else {
    "other"
  }
}

infer_role <- function(name, type, missing_rate, unique_rate, n_unique, n_rows, values) {
  nm <- tolower(name)
  if (grepl("(^|_)(target|label|outcome|response|class|churn|status)($|_)", nm)) {
    return("target")
  }
  if (type == "date" || grepl("date|time|timestamp|created|updated|period", nm)) {
    return("time")
  }
  if (type == "logical" || grepl("(^is_|^has_|flag$|_flag$)", nm)) {
    return("flag")
  }
  id_like_name <- grepl("(^id$|_id$|uuid|guid|key$|code$|email)", nm)
  id_like_values <- type %in% c("character", "factor") &&
    is.finite(unique_rate) &&
    unique_rate > 0.95 &&
    n_unique >= min(20L, max(1L, floor(0.8 * n_rows)))
  if (missing_rate < 0.05 && (id_like_name || id_like_values)) {
    return("id")
  }
  if (type == "numeric") {
    return("measure")
  }
  if (type %in% c("character", "factor")) {
    lengths <- nchar(as.character(values), allowNA = FALSE)
    if (length(lengths) && stats::median(lengths) > 35 && unique_rate > 0.2) {
      return("text")
    }
    return("category")
  }
  "unknown"
}

category_entropy <- function(x) {
  if (!length(x)) {
    return(NA_real_)
  }
  p <- table(x) / length(x)
  -sum(p * log(p)) / log(length(p) + 1)
}

category_tail_rate <- function(x) {
  if (!length(x)) {
    return(NA_real_)
  }
  freq <- sort(table(x), decreasing = TRUE)
  if (length(freq) <= 5L) {
    return(0)
  }
  sum(freq[-seq_len(5L)]) / sum(freq)
}

top_values <- function(x, n = 5L) {
  if (!length(x)) {
    return(character())
  }
  names(sort(table(x), decreasing = TRUE))[seq_len(min(n, length(unique(x))))]
}

dna_species_from_profile <- function(columns, n_rows, n_cols) {
  if (!n_cols) {
    return("unknown_table")
  }
  names_lower <- tolower(columns$name)
  roles <- columns$role
  numeric_cols <- sum(columns$type == "numeric")
  category_cols <- sum(columns$role == "category")
  time_cols <- sum(columns$role == "time")
  id_cols <- sum(columns$role == "id")

  if (n_cols >= 50L && numeric_cols / n_cols >= 0.6) {
    "wide_feature_matrix"
  } else if (any(grepl("customer|client|user|account", names_lower)) && id_cols >= 1L) {
    "customer_table"
  } else if (any(grepl("amount|price|payment|order|invoice|transaction|qty|quantity", names_lower))) {
    "transaction_log"
  } else if (time_cols >= 1L && n_rows / max(n_cols, 1L) > 20 && (id_cols >= 1L || category_cols >= 1L)) {
    "event_stream"
  } else if (time_cols >= 1L && n_rows >= 10L && n_cols <= 8L && numeric_cols >= 1L) {
    "time_series"
  } else if (category_cols / max(n_cols, 1L) >= 0.6 && n_cols >= 5L) {
    "survey_data"
  } else if (id_cols >= 1L && n_rows <= 5000L && category_cols >= numeric_cols) {
    "lookup_table"
  } else {
    "unknown_table"
  }
}

dna_scores <- function(columns, n_rows, n_cols) {
  missing_penalty <- safe_mean(columns$missing_rate)
  unknown_penalty <- mean(columns$role == "unknown")
  id_bonus <- min(0.15, sum(columns$role == "id") * 0.05)
  identity_strength <- 100 * clip01(0.72 - 0.45 * missing_penalty - 0.2 * unknown_penalty + id_bonus + 0.13 * min(1, n_cols / 12))
  risk_value <- clip01(0.55 * missing_penalty + 0.25 * mean(columns$unique_rate < 0.02, na.rm = TRUE) + 0.2 * mean(columns$outlier_rate > 0.08, na.rm = TRUE))
  stability_value <- clip01(1 - risk_value)
  list(
    identity_strength = identity_strength,
    stability_value = stability_value,
    stability = if (stability_value >= 0.8) "high" else if (stability_value >= 0.55) "medium" else "low",
    drift_risk_value = risk_value,
    drift_risk = if (risk_value >= 0.65) "high" else if (risk_value >= 0.35) "medium" else "low"
  )
}

make_fingerprints <- function(columns, n_rows, n_cols) {
  schema <- paste(columns$name, columns$type, sep = ":")
  shape <- paste(floor(log10(max(n_rows, 1L))), n_cols, sep = "x")
  profile <- paste(
    columns$name,
    columns$role,
    round_num(columns$missing_rate, 2),
    round_num(columns$unique_rate, 2),
    round_num(columns$q50, 2),
    round_num(columns$entropy, 2),
    sep = ":"
  )
  list(
    schema_hash = stable_hash(schema),
    shape_hash = stable_hash(shape),
    profile_hash = stable_hash(profile)
  )
}

dna_signals <- function(columns, n_rows, n_cols, species) {
  role_counts <- sort(table(columns$role), decreasing = TRUE)
  signals <- c(
    paste0("species=", species),
    paste0(n_cols, " columns profiled"),
    paste0(sum(columns$role == "id"), " likely ID columns"),
    paste0(sum(columns$role == "measure"), " numeric measure columns"),
    paste0(sum(columns$role == "category"), " categorical columns")
  )
  if (length(role_counts)) {
    signals <- c(signals, paste0("dominant role=", names(role_counts)[[1L]]))
  }
  signals
}

dna_warnings <- function(columns) {
  warnings <- character()
  high_missing <- columns$name[columns$missing_rate >= 0.3]
  high_outliers <- columns$name[columns$outlier_rate >= 0.1 & !is.na(columns$outlier_rate)]
  high_tail <- columns$name[columns$role == "category" & columns$tail_rate >= 0.6 & !is.na(columns$tail_rate)]
  no_id <- !any(columns$role == "id")

  if (length(high_missing)) warnings <- c(warnings, paste0("high missingness: ", paste(head(high_missing, 3L), collapse = ", ")))
  if (length(high_outliers)) warnings <- c(warnings, paste0("outlier-heavy measures: ", paste(head(high_outliers, 3L), collapse = ", ")))
  if (length(high_tail)) warnings <- c(warnings, paste0("long-tail categories: ", paste(head(high_tail, 3L), collapse = ", ")))
  if (no_id) warnings <- c(warnings, "no strong ID column detected")
  warnings
}

#' Match a data set against a DNA library
#'
#' Finds the closest relatives of a query data set by comparing its data DNA
#' against a named library of data frames or `data_dna` objects.
#'
#' @param x A data frame or `data_dna` object to match.
#' @param library A list of data frames or `data_dna` objects.
#' @param top_n Maximum number of matches to return.
#' @param sample_size Maximum number of rows used when profiling raw data frames.
#'
#' @return A `dna_match` object.
#' @export
#'
#' @examples
#' demo <- dna_example_customers()
#' lib <- list(old = data_dna(demo$customers_old), new = data_dna(demo$customers_new))
#' dna_match(demo$customers_new, lib)
dna_match <- function(x, library, top_n = 5L, sample_size = 10000L) {
  if (!is.list(library)) {
    stop("`library` must be a list of data frames or data_dna objects.", call. = FALSE)
  }
  if (!length(library)) {
    stop("`library` must contain at least one candidate data set.", call. = FALSE)
  }
  stopifnot(length(top_n) == 1L, top_n >= 1L)

  query <- as_data_dna(x, name = deparse(substitute(x)), sample_size = sample_size)
  candidate_names <- match_library_names(library)
  comparisons <- vector("list", length(library))
  rows <- vector("list", length(library))

  for (i in seq_along(library)) {
    candidate <- as_data_dna(library[[i]], name = candidate_names[[i]], sample_size = sample_size)
    candidate$name <- candidate_names[[i]]
    comparison <- dna_compare(query, candidate)
    comparisons[[i]] <- comparison
    rows[[i]] <- match_row(candidate_names[[i]], candidate, comparison)
  }

  matches <- do.call(rbind, rows)
  matches$is_self <- matches$dna_id == query$dna_id
  order_idx <- order(matches$is_self, -matches$overall, na.last = TRUE)
  matches <- matches[order_idx, , drop = FALSE]
  comparisons <- comparisons[order_idx]

  actual_top_n <- min(as.integer(top_n), nrow(matches))
  matches <- matches[seq_len(actual_top_n), , drop = FALSE]
  comparisons <- comparisons[seq_len(actual_top_n)]
  row.names(matches) <- NULL

  structure(
    list(
      query = query,
      matches = matches,
      comparisons = comparisons,
      top_n = actual_top_n
    ),
    class = "dna_match"
  )
}

#' Render a DataDNA lineage match card
#'
#' Creates a static HTML/SVG lineage diagram for a `dna_match` object.
#'
#' @param match A `dna_match` object.
#' @param file Optional HTML file path. If supplied, the card is saved there.
#' @param open Logical. Open the saved file in a browser when `file` is supplied.
#'
#' @return An `htmltools` browsable object, invisibly when saved to `file`.
#' @export
#'
#' @examples
#' demo <- dna_example_customers()
#' lib <- list(old = data_dna(demo$customers_old), new = data_dna(demo$customers_new))
#' match <- dna_match(demo$customers_new, lib)
#' dna_match_card(match)
dna_match_card <- function(match, file = NULL, open = FALSE) {
  if (!inherits(match, "dna_match")) {
    stop("`match` must be a dna_match object.", call. = FALSE)
  }

  card <- build_match_card(match)
  if (!is.null(file)) {
    htmltools::save_html(card, file = file)
    if (isTRUE(open)) {
      utils::browseURL(normalizePath(file, winslash = "/", mustWork = FALSE))
    }
    return(invisible(card))
  }

  htmltools::browsable(card)
}

#' @export
print.dna_match <- function(x, ...) {
  cat("DataDNA lineage match\n")
  cat("  Query: ", x$query$name, " (", x$query$dna_id, ")\n", sep = "")
  if (!nrow(x$matches)) {
    cat("  No matches.\n")
    return(invisible(x))
  }

  best <- x$matches[1L, , drop = FALSE]
  cat("  Best match: ", best$name, " | ", round(best$overall, 3), " | ", best$relation, "\n", sep = "")
  cat("  Top candidates:\n")
  for (i in seq_len(nrow(x$matches))) {
    cat(
      "    ",
      i,
      ". ",
      x$matches$name[[i]],
      " (",
      round(x$matches$overall[[i]], 3),
      ", ",
      x$matches$relation[[i]],
      ")\n",
      sep = ""
    )
  }
  notes <- x$matches$top_mutations[[1L]]
  if (length(notes)) {
    cat("  Best-match mutations: ", paste(notes, collapse = " | "), "\n", sep = "")
  }
  invisible(x)
}

match_library_names <- function(library) {
  labels <- names(library)
  if (is.null(labels)) {
    labels <- rep("", length(library))
  }

  empty <- !nzchar(labels) | is.na(labels)
  if (any(empty)) {
    labels[empty] <- vapply(
      which(empty),
      function(i) {
        item <- library[[i]]
        if (inherits(item, "data_dna") && length(item$name) == 1L && nzchar(item$name)) {
          item$name
        } else {
          paste0("candidate_", i)
        }
      },
      character(1)
    )
  }

  make.unique(labels, sep = "_")
}

match_row <- function(name, candidate, comparison) {
  data.frame(
    name = name,
    dna_id = candidate$dna_id,
    species = candidate$species,
    overall = comparison$overall,
    relation = comparison$likely_relation,
    schema = unname(comparison$scores[["schema"]]),
    shape = unname(comparison$scores[["shape"]]),
    species_score = unname(comparison$scores[["species"]]),
    role_structure = unname(comparison$scores[["role_structure"]]),
    distribution = unname(comparison$scores[["distribution"]]),
    missingness = unname(comparison$scores[["missingness"]]),
    category = unname(comparison$scores[["category"]]),
    identity = unname(comparison$scores[["identity"]]),
    top_mutations = I(list(comparison$top_mutations)),
    stringsAsFactors = FALSE
  )
}

build_match_card <- function(match) {
  best <- match$matches[1L, , drop = FALSE]
  htmltools::tags$html(
    htmltools::tags$head(
      htmltools::tags$meta(charset = "utf-8"),
      htmltools::tags$title(paste("DataDNA lineage", match$query$name)),
      htmltools::tags$style(htmltools::HTML(match_card_css()))
    ),
    htmltools::tags$body(
      htmltools::tags$main(
        class = "dna-shell",
        htmltools::tags$section(
          class = "dna-card lineage-report",
          htmltools::tags$div(
            class = "report-header",
            htmltools::tags$div(
              class = "report-title",
              htmltools::tags$div(class = "kicker", "DataDNA lineage report"),
              htmltools::tags$h1(match$query$name),
              htmltools::tags$p(class = "dna-id", match$query$dna_id),
              htmltools::tags$div(
                class = "report-tags",
                htmltools::tags$span(paste("Species", match$query$species, sep = ": ")),
                htmltools::tags$span(paste("Rows", format(match$query$n_rows, big.mark = ","), sep = ": ")),
                htmltools::tags$span(paste("Columns", match$query$n_cols, sep = ": "))
              )
            ),
            htmltools::tags$div(
              class = "best-card",
              htmltools::tags$span("Closest relative"),
              htmltools::tags$strong(best$name),
              htmltools::tags$div(
                class = "best-score",
                htmltools::tags$em(format_pct(best$overall)),
                htmltools::tags$b(best$relation)
              )
            )
          ),
          htmltools::tags$div(
            class = "report-metrics",
            lineage_stat("Relation", best$relation, "relation"),
            lineage_stat("Similarity", format_pct(best$overall), "score"),
            lineage_stat("Best species", best$species, "species"),
            lineage_stat("Candidates scanned", nrow(match$matches), "count")
          ),
          htmltools::tags$div(
            class = "report-body",
            htmltools::tags$div(
              class = "lineage-panel lineage-graph-panel",
              htmltools::tags$div(
                class = "panel-heading",
                htmltools::tags$span("Lineage map"),
                htmltools::tags$h2("Similarity-weighted candidate graph")
              ),
              htmltools::HTML(lineage_svg(match))
            ),
            htmltools::tags$div(
              class = "lineage-panel lineage-rank",
              htmltools::tags$div(
                class = "panel-heading",
                htmltools::tags$span("Ranked relatives"),
                htmltools::tags$h2("Candidate ranking")
              ),
              match_table(match)
            )
          ),
          htmltools::tags$div(
            class = "lineage-summary",
            htmltools::tags$div(
              class = "panel-heading",
              htmltools::tags$span("Interpretation"),
              htmltools::tags$h2("Mutation summary")
            ),
            mutation_summary(match)
          )
        )
      )
    )
  )
}

lineage_stat <- function(label, value, kind) {
  htmltools::tags$div(
    class = paste("lineage-stat", kind),
    htmltools::tags$span(label),
    htmltools::tags$strong(value)
  )
}

short_label <- function(x, width = 26L) {
  x <- as.character(x)
  ifelse(nchar(x, type = "width") > width, paste0(substr(x, 1L, width - 1L), "..."), x)
}

lineage_svg <- function(match) {
  matches <- match$matches
  n <- nrow(matches)
  width <- 960
  height <- max(330, 130 + n * 92)
  query_x <- 72
  query_y <- height / 2
  query_w <- 250
  query_h <- 104
  cand_x <- 606
  cand_w <- 292
  cand_h <- 78
  cand_y <- if (n == 1L) height / 2 else seq(86, height - 86, length.out = n)

  lines <- character(n)
  nodes <- character(n)
  for (i in seq_len(n)) {
    score <- matches$overall[[i]]
    color <- relation_color(score)
    stroke <- 1.4 + 3.4 * clip01(score)
    label <- htmltools::htmlEscape(short_label(matches$name[[i]], 24L))
    relation <- htmltools::htmlEscape(matches$relation[[i]])
    pct <- format_pct(score)
    y_top <- cand_y[[i]] - cand_h / 2
    lines[[i]] <- sprintf(
      "<path d='M %s %s C %s %s, %s %s, %s %s' stroke='%s' stroke-width='%.1f' fill='none' opacity='.86'/>",
      query_x + query_w,
      query_y,
      query_x + query_w + 88,
      query_y,
      cand_x - 94,
      cand_y[[i]],
      cand_x,
      cand_y[[i]],
      color,
      stroke
    )
    nodes[[i]] <- sprintf(
      paste0(
        "<g>",
        "<rect x='%s' y='%s' width='%s' height='%s' rx='12' fill='#111b25' stroke='%s' stroke-width='1.2'/>",
        "<rect x='%s' y='%s' width='4' height='%s' rx='2' fill='%s'/>",
        "<text x='%s' y='%s' class='node-title'>%s</text>",
        "<text x='%s' y='%s' class='node-sub'>%s</text>",
        "<text x='%s' y='%s' text-anchor='end' class='node-score'>%s</text>",
        "</g>"
      ),
      cand_x,
      y_top,
      cand_w,
      cand_h,
      color,
      cand_x,
      y_top + 16,
      cand_h - 32,
      color,
      cand_x + 22,
      cand_y[[i]] - 8,
      label,
      cand_x + 22,
      cand_y[[i]] + 14,
      relation,
      cand_x + cand_w - 20,
      cand_y[[i]] + 2,
      pct
    )
  }

  query_label <- htmltools::htmlEscape(short_label(match$query$name, 22L))
  sprintf(
    paste0(
      "<svg class='lineage-svg' viewBox='0 0 %s %s' role='img' aria-label='DataDNA lineage graph'>",
      "<style>",
      ".node-title{font:600 14px Inter,Segoe UI,Helvetica Neue,Arial,sans-serif;fill:#eef3f6;}",
      ".node-score{font:700 20px Inter,Segoe UI,Helvetica Neue,Arial,sans-serif;fill:#eef3f6;}",
      ".node-sub{font:12px Inter,Segoe UI,Helvetica Neue,Arial,sans-serif;fill:#9aa8b5;}",
      ".query-title{font:700 16px Inter,Segoe UI,Helvetica Neue,Arial,sans-serif;fill:#eef3f6;}",
      ".query-sub{font:12px SFMono-Regular,Consolas,Liberation Mono,monospace;fill:#9aa8b5;}",
      ".section-label{font:600 11px Inter,Segoe UI,Helvetica Neue,Arial,sans-serif;fill:#6f7a86;}",
      "</style>",
      "<defs>",
      "<pattern id='grid' width='40' height='40' patternUnits='userSpaceOnUse'>",
      "<path d='M 40 0 L 0 0 0 40' fill='none' stroke='rgba(255,255,255,.035)' stroke-width='1'/>",
      "</pattern>",
      "</defs>",
      "<rect width='100%%' height='100%%' rx='18' fill='#0f1720'/>",
      "<rect width='100%%' height='100%%' rx='18' fill='url(#grid)'/>",
      "<text x='%s' y='42' class='section-label'>QUERY SPECIMEN</text>",
      "<text x='%s' y='42' class='section-label'>RELATIVE CANDIDATES</text>",
      "%s",
      "<g>",
      "<rect x='%s' y='%s' width='%s' height='%s' rx='14' fill='#121d28' stroke='#66d9c2' stroke-width='1.4'/>",
      "<rect x='%s' y='%s' width='4' height='%s' rx='2' fill='#66d9c2'/>",
      "<text x='%s' y='%s' class='query-title'>%s</text>",
      "<text x='%s' y='%s' class='query-sub'>%s</text>",
      "<text x='%s' y='%s' class='node-sub'>%s</text>",
      "</g>",
      "%s",
      "</svg>"
    ),
    width,
    height,
    query_x,
    cand_x,
    paste(lines, collapse = ""),
    query_x,
    query_y - query_h / 2,
    query_w,
    query_h,
    query_x,
    query_y - query_h / 2 + 18,
    query_h - 36,
    query_x + 24,
    query_y - 14,
    query_label,
    query_x + 24,
    query_y + 10,
    match$query$dna_id,
    query_x + 24,
    query_y + 34,
    match$query$species,
    paste(nodes, collapse = "")
  )
}

match_table <- function(match) {
  htmltools::tags$div(
    class = "rank-list",
    lapply(seq_len(nrow(match$matches)), function(i) {
      mutations <- match$matches$top_mutations[[i]]
      score <- match$matches$overall[[i]]
      htmltools::tags$article(
        class = "rank-card",
        htmltools::tags$div(
          class = "rank-topline",
          htmltools::tags$span(class = "rank-index", sprintf("%02d", i)),
          htmltools::tags$strong(match$matches$name[[i]]),
          htmltools::tags$em(format_pct(score))
        ),
        htmltools::tags$div(
          class = "rank-meter",
          htmltools::tags$i(style = paste0("width:", round(100 * clip01(score)), "%; background:", relation_color(score), ";"))
        ),
        htmltools::tags$p(match$matches$relation[[i]]),
        htmltools::tags$small(if (length(mutations)) paste(utils::head(mutations, 2L), collapse = " | ") else "No notable mutations")
      )
    })
  )
}

mutation_summary <- function(match) {
  best <- match$matches[1L, , drop = FALSE]
  notes <- best$top_mutations[[1L]]
  if (!length(notes)) {
    notes <- "No notable mutations detected for the closest relative."
  }

  htmltools::tags$div(
    class = "summary-grid",
    htmltools::tags$div(
      class = "summary-callout",
      htmltools::tags$span("Best lineage interpretation"),
      htmltools::tags$strong(best$relation),
      htmltools::tags$p(paste0(best$name, " is the nearest non-self relative with ", format_pct(best$overall), " similarity."))
    ),
    htmltools::tags$ul(
      class = "summary-notes",
      lapply(utils::head(notes, 4L), htmltools::tags$li)
    )
  )
}

relation_label_class <- function(score) {
  if (score >= 0.97) {
    "near"
  } else if (score >= 0.85) {
    "changed"
  } else if (score >= 0.65) {
    "related"
  } else {
    "different"
  }
}

relation_color <- function(score) {
  if (score >= 0.97) {
    "#66d9c2"
  } else if (score >= 0.85) {
    "#69a7d8"
  } else if (score >= 0.65) {
    "#d8b85f"
  } else {
    "#6f7a86"
  }
}

risk_from_similarity <- function(score) {
  if (score >= 0.85) "low" else if (score >= 0.65) "medium" else "high"
}

match_card_css <- function() {
  paste0(
    dna_card_css(),
    "
    body {
      color: #eef3f6;
      background:
        radial-gradient(circle at 18% 0%, rgba(102, 217, 194, .08), transparent 360px),
        linear-gradient(180deg, #0b1118 0%, #101923 100%);
      font-family: Inter, 'Segoe UI', 'Helvetica Neue', Arial, sans-serif;
    }
    .dna-shell {
      width: min(1180px, calc(100vw - 40px));
      margin: 28px auto;
    }
    .lineage-report {
      border: 1px solid rgba(255,255,255,.10);
      border-radius: 8px;
      background: #0f1720;
      box-shadow: 0 28px 90px rgba(0,0,0,.34);
    }
    .report-header {
      display: grid;
      grid-template-columns: minmax(0, 1fr) 320px;
      gap: 28px;
      padding: 34px 36px 30px;
      border-bottom: 1px solid rgba(255,255,255,.10);
      background: linear-gradient(180deg, rgba(255,255,255,.035), rgba(255,255,255,.01));
    }
    .report-title h1 {
      margin: 10px 0 8px;
      color: #eef3f6;
      font-size: 40px;
      line-height: 1.05;
      font-weight: 700;
      letter-spacing: 0;
    }
    .kicker {
      color: #66d9c2;
      font-size: 12px;
      font-weight: 700;
      text-transform: uppercase;
      letter-spacing: 0;
    }
    .dna-id {
      margin: 0;
      color: #9aa8b5;
      font-family: SFMono-Regular, Consolas, 'Liberation Mono', monospace;
      font-size: 14px;
    }
    .report-tags {
      display: flex;
      flex-wrap: wrap;
      gap: 8px;
      margin-top: 20px;
    }
    .report-tags span {
      padding: 7px 10px;
      border: 1px solid rgba(255,255,255,.10);
      border-radius: 6px;
      color: #cdd6de;
      background: rgba(255,255,255,.035);
      font-size: 12px;
    }
    .best-card {
      align-self: stretch;
      display: grid;
      align-content: center;
      gap: 12px;
      padding: 24px;
      border: 1px solid rgba(102,217,194,.28);
      border-radius: 8px;
      background: #101923;
    }
    .best-card > span {
      color: #9aa8b5;
      font-size: 12px;
      font-weight: 700;
      text-transform: uppercase;
    }
    .best-card > strong {
      color: #eef3f6;
      font-size: 28px;
      line-height: 1.08;
      font-weight: 700;
    }
    .best-score {
      display: flex;
      align-items: flex-end;
      gap: 12px;
    }
    .best-score em {
      color: #66d9c2;
      font-size: 40px;
      line-height: .95;
      font-style: normal;
      font-weight: 700;
    }
    .best-score b {
      color: #9aa8b5;
      font-size: 12px;
      font-weight: 600;
      line-height: 1.25;
      padding-bottom: 4px;
    }
    .report-metrics {
      display: grid;
      grid-template-columns: 1.35fr .7fr .9fr .65fr;
      gap: 1px;
      background: rgba(255,255,255,.10);
    }
    .lineage-stat {
      min-width: 0;
      padding: 18px 24px;
      background: #101923;
    }
    .lineage-stat span {
      display: block;
      color: #9aa8b5;
      font-size: 12px;
      font-weight: 650;
      text-transform: uppercase;
    }
    .lineage-stat strong {
      display: block;
      margin-top: 7px;
      color: #eef3f6;
      font-size: 16px;
      line-height: 1.25;
      font-weight: 650;
    }
    .report-body {
      display: grid;
      grid-template-columns: minmax(0, 1.45fr) minmax(320px, .75fr);
      gap: 20px;
      padding: 24px;
    }
    .lineage-panel,
    .lineage-summary {
      border: 1px solid rgba(255,255,255,.10);
      border-radius: 8px;
      background: rgba(255,255,255,.025);
    }
    .lineage-panel {
      padding: 20px;
    }
    .panel-heading {
      margin-bottom: 16px;
    }
    .panel-heading span {
      display: block;
      color: #66d9c2;
      font-size: 12px;
      font-weight: 700;
      text-transform: uppercase;
    }
    .panel-heading h2 {
      margin: 6px 0 0;
      color: #eef3f6;
      font-size: 20px;
      line-height: 1.2;
      font-weight: 650;
      letter-spacing: 0;
    }
    .lineage-svg {
      display: block;
      width: 100%;
      height: auto;
      min-height: 330px;
    }
    .rank-list {
      display: grid;
      gap: 12px;
    }
    .rank-card {
      padding: 14px;
      border: 1px solid rgba(255,255,255,.10);
      border-radius: 8px;
      background: #111b25;
    }
    .rank-topline {
      display: grid;
      grid-template-columns: 34px minmax(0, 1fr) auto;
      align-items: baseline;
      gap: 10px;
    }
    .rank-index {
      color: #6f7a86;
      font-family: SFMono-Regular, Consolas, 'Liberation Mono', monospace;
      font-size: 12px;
    }
    .rank-topline strong {
      min-width: 0;
      color: #eef3f6;
      font-size: 16px;
      font-weight: 650;
      overflow-wrap: anywhere;
    }
    .rank-topline em {
      color: #66d9c2;
      font-size: 16px;
      font-style: normal;
      font-weight: 700;
    }
    .rank-meter {
      height: 4px;
      margin: 11px 0 10px;
      border-radius: 999px;
      background: rgba(255,255,255,.08);
      overflow: hidden;
    }
    .rank-meter i {
      display: block;
      height: 100%;
      border-radius: inherit;
    }
    .rank-card p {
      margin: 0 0 7px;
      color: #cdd6de;
      font-size: 14px;
      line-height: 1.35;
    }
    .rank-card small {
      color: #9aa8b5;
      font-size: 12px;
      line-height: 1.35;
    }
    .lineage-summary {
      margin: 0 24px 24px;
      padding: 20px;
    }
    .summary-grid {
      display: grid;
      grid-template-columns: minmax(0, .9fr) minmax(0, 1.1fr);
      gap: 20px;
    }
    .summary-callout {
      padding: 16px;
      border-radius: 8px;
      background: #101923;
      border: 1px solid rgba(255,255,255,.10);
    }
    .summary-callout span {
      color: #9aa8b5;
      font-size: 12px;
      font-weight: 700;
      text-transform: uppercase;
    }
    .summary-callout strong {
      display: block;
      margin-top: 8px;
      color: #eef3f6;
      font-size: 20px;
      line-height: 1.2;
    }
    .summary-callout p {
      margin: 10px 0 0;
      color: #9aa8b5;
      font-size: 14px;
      line-height: 1.45;
    }
    .summary-notes {
      margin: 0;
      padding: 0;
      list-style: none;
      display: grid;
      gap: 8px;
    }
    .summary-notes li {
      padding: 11px 13px;
      border-radius: 6px;
      color: #cdd6de;
      background: rgba(255,255,255,.035);
      border: 1px solid rgba(255,255,255,.08);
      font-size: 14px;
      line-height: 1.35;
    }
    @media (max-width: 920px) {
      .report-header,
      .report-body,
      .summary-grid {
        grid-template-columns: 1fr;
      }
      .report-metrics {
        grid-template-columns: repeat(2, 1fr);
      }
    }
    "
  )
}

library(DataDNA)

demo <- dna_example_customers()
stopifnot(is.data.frame(customers_old))
stopifnot(is.data.frame(customers_new))

dna <- data_dna(demo$customers_new, name = "customers_new")
stopifnot(inherits(dna, "data_dna"))
stopifnot(grepl("^DNA-", dna$dna_id))
stopifnot(dna$n_rows == nrow(demo$customers_new))
stopifnot(any(dna$columns$role == "id"))
stopifnot(dna_species(demo$customers_new) %in% c("customer_table", "unknown_table"))

shuffled <- demo$customers_new[sample(seq_len(nrow(demo$customers_new))), ]
same <- dna_compare(demo$customers_new, shuffled)
stopifnot(same$overall > 0.95)

changed <- dna_compare(demo$customers_old, demo$customers_new)
stopifnot(changed$overall < 1)
stopifnot(changed$overall > 0.5)

mutations <- dna_diff(demo$customers_old, demo$customers_new)
stopifnot(inherits(mutations, "dna_diff"))
stopifnot(any(grepl("loyalty_score", mutations$changes$message)))
stopifnot(any(mutations$changes$type %in% c("missingness", "distribution", "category")))

tmp <- tempfile(fileext = ".html")
invisible(dna_card(dna, file = tmp))
html <- paste(readLines(tmp, warn = FALSE), collapse = "\n")
stopifnot(file.exists(tmp))
stopifnot(grepl("DataDNA specimen card", html))
stopifnot(grepl(dna$dna_id, html, fixed = TRUE))
stopifnot(grepl("fingerprint-chip", html))
stopifnot(grepl("legend", html))

set.seed(101)
n <- 500
customer_base <- data.frame(
  customer_id = sprintf("C%05d", seq_len(n)),
  signup_date = as.Date("2025-01-01") + sample(0:365, n, TRUE),
  region = sample(c("North", "South", "East", "West"), n, TRUE, c(.30, .25, .25, .20)),
  plan = sample(c("Free", "Pro", "Team", "Enterprise"), n, TRUE, c(.52, .28, .15, .05)),
  age = round(rnorm(n, 39, 10)),
  spend = round(rlnorm(n, log(900), .55), 2),
  visits = rpois(n, 12),
  churn_target = sample(c("no", "yes"), n, TRUE, c(.82, .18)),
  stringsAsFactors = FALSE
)
customer_mutated <- customer_base
customer_mutated$spend <- round(customer_mutated$spend * runif(n, 1.12, 1.45), 2)
customer_mutated$region[sample(seq_len(n), 70)] <- "International"
customer_mutated$age[sample(seq_len(n), 90)] <- NA
customer_mutated$visits <- customer_mutated$visits + rpois(n, 4)
customer_mutated$loyalty_score <- round(runif(n, 0, 100), 1)

sensor_stream <- data.frame(
  sensor_id = sprintf("S%03d", sample(1:90, n, TRUE)),
  timestamp = as.POSIXct("2026-01-01", tz = "UTC") + sort(sample(1:900000, n)),
  temperature = rnorm(n, 22, 3),
  vibration = rgamma(n, 2, .8),
  pressure = rnorm(n, 101, 2),
  fault_flag = sample(c(TRUE, FALSE), n, TRUE, c(.04, .96)),
  batch = sample(LETTERS[1:12], n, TRUE),
  stringsAsFactors = FALSE
)

base_vs_mutated <- dna_compare(customer_base, customer_mutated)
base_vs_sensor <- dna_compare(customer_base, sensor_stream)
stress_diff <- dna_diff(customer_base, customer_mutated)

stopifnot(dna_species(customer_base) == "customer_table")
stopifnot(dna_species(customer_mutated) == "customer_table")
stopifnot(dna_species(sensor_stream) == "event_stream")
stopifnot(base_vs_mutated$overall > 0.8)
stopifnot(base_vs_mutated$likely_relation == "same dataset, changed version")
stopifnot(all(c("shape", "species", "role_structure") %in% names(base_vs_mutated$scores)))
stopifnot(base_vs_mutated$scores[["species"]] == 1)
stopifnot(length(base_vs_mutated$top_mutations) >= 1)
stopifnot(base_vs_sensor$overall < 0.25)
stopifnot(base_vs_sensor$likely_relation == "different dataset")
stopifnot(base_vs_sensor$scores[["schema"]] == 0)
stopifnot(base_vs_sensor$scores[["species"]] == 0)
stopifnot(any(grepl("loyalty_score", stress_diff$changes$message)))
stopifnot(any(grepl("spend", stress_diff$changes$message)))
stopifnot(any(grepl("visits", stress_diff$changes$message)))

lineage_library <- list(
  customer_base = customer_base,
  customer_mutated = customer_mutated,
  sensor_stream = sensor_stream
)
lineage <- dna_match(customer_mutated, lineage_library, top_n = 10)
stopifnot(inherits(lineage, "dna_match"))
stopifnot(nrow(lineage$matches) == 3)
stopifnot(lineage$matches$name[[1L]] == "customer_base")
stopifnot(lineage$matches$overall[[1L]] > lineage$matches$overall[lineage$matches$name == "sensor_stream"])
stopifnot(lineage$matches$relation[[1L]] %in% c("same dataset, changed version", "same dataset, near-identical version"))

unnamed_lineage <- dna_match(customer_mutated, list(customer_base, sensor_stream), top_n = 10)
stopifnot(all(grepl("^candidate_", unnamed_lineage$matches$name)))
stopifnot(nrow(unnamed_lineage$matches) == 2)

too_many <- dna_match(customer_mutated, list(base = customer_base), top_n = 99)
stopifnot(nrow(too_many$matches) == 1)
stopifnot(too_many$top_n == 1)

empty_error <- tryCatch(
  dna_match(customer_mutated, list()),
  error = function(e) conditionMessage(e)
)
stopifnot(grepl("at least one", empty_error))

lineage_html_file <- tempfile(fileext = ".html")
invisible(dna_match_card(lineage, file = lineage_html_file))
lineage_html <- paste(readLines(lineage_html_file, warn = FALSE), collapse = "\n")
stopifnot(file.exists(lineage_html_file))
stopifnot(grepl("<svg", lineage_html, fixed = TRUE))
stopifnot(grepl(lineage$query$dna_id, lineage_html, fixed = TRUE))
stopifnot(grepl("customer_base", lineage_html, fixed = TRUE))
stopifnot(grepl("Similarity", lineage_html, fixed = TRUE))
stopifnot(grepl("lineage-report", lineage_html, fixed = TRUE))
stopifnot(grepl("lineage-rank", lineage_html, fixed = TRUE))
stopifnot(grepl("lineage-summary", lineage_html, fixed = TRUE))

lineage_png_file <- tempfile(fileext = ".png")
invisible(dna_match_plot(lineage, file = lineage_png_file, width = 9, height = 6))
stopifnot(file.exists(lineage_png_file))
stopifnot(file.info(lineage_png_file)$size > 1000)

lineage_pdf_file <- tempfile(fileext = ".pdf")
invisible(dna_match_plot(lineage, file = lineage_pdf_file, width = 9, height = 6))
stopifnot(file.exists(lineage_pdf_file))
stopifnot(file.info(lineage_pdf_file)$size > 1000)

lineage_large_png_file <- tempfile(fileext = ".png")
invisible(dna_match_plot(lineage, file = lineage_large_png_file, width = 11, height = 7))
stopifnot(file.exists(lineage_large_png_file))
stopifnot(file.info(lineage_large_png_file)$size > 1000)

long_name_lineage <- dna_match(
  customer_mutated,
  list(
    customer_base_2024_archive_version_with_extra_long_name = customer_base,
    sensor_stream_operational_reference_with_extra_long_name = sensor_stream
  ),
  top_n = 2
)
long_name_png_file <- tempfile(fileext = ".png")
invisible(dna_match_plot(long_name_lineage, file = long_name_png_file, width = 9, height = 6))
stopifnot(file.exists(long_name_png_file))
stopifnot(file.info(long_name_png_file)$size > 1000)

single_match <- dna_match(customer_mutated, list(customer_base = customer_base), top_n = 1)
single_match_png_file <- tempfile(fileext = ".png")
invisible(dna_match_plot(single_match, file = single_match_png_file, width = 9, height = 6))
stopifnot(file.exists(single_match_png_file))
stopifnot(file.info(single_match_png_file)$size > 1000)

as_data_dna <- function(x, name = NULL, sample_size = 10000L) {
  if (inherits(x, "data_dna")) {
    return(x)
  }
  data_dna(x, name = name, sample_size = sample_size)
}

stable_hash <- function(x) {
  text <- paste(as.character(x), collapse = "\r")
  bytes <- as.integer(charToRaw(enc2utf8(text)))
  hash <- 2166136261
  modulus <- 4294967291

  if (!length(bytes)) {
    return("00000000")
  }

  for (byte in bytes) {
    hash <- (hash + byte) %% modulus
    hash <- (hash * 16777619) %% modulus
  }

  sprintf("%08X", as.integer(hash %% 2147483647))
}

compact_id <- function(...) {
  raw <- stable_hash(unlist(list(...), use.names = FALSE))
  paste0("DNA-", substr(raw, 1, 4), "-", substr(raw, 5, 8), "-", substr(stable_hash(raw), 1, 4))
}

safe_mean <- function(x) {
  x <- x[is.finite(x)]
  if (!length(x)) NA_real_ else mean(x)
}

safe_sd <- function(x) {
  x <- x[is.finite(x)]
  if (length(x) < 2L) NA_real_ else stats::sd(x)
}

safe_quantile <- function(x) {
  x <- x[is.finite(x)]
  if (!length(x)) {
    return(rep(NA_real_, 5L))
  }
  as.numeric(stats::quantile(x, probs = c(0, 0.25, 0.5, 0.75, 1), names = FALSE, type = 7))
}

clip01 <- function(x) {
  pmax(0, pmin(1, x))
}

jaccard <- function(x, y) {
  x <- unique(x[!is.na(x)])
  y <- unique(y[!is.na(y)])
  denom <- length(union(x, y))
  if (denom == 0L) {
    return(1)
  }
  length(intersect(x, y)) / denom
}

weighted_score <- function(values, weights) {
  ok <- is.finite(values) & is.finite(weights) & weights > 0
  if (!any(ok)) {
    return(NA_real_)
  }
  sum(values[ok] * weights[ok]) / sum(weights[ok])
}

format_pct <- function(x, digits = 0L) {
  if (!is.finite(x)) {
    return("NA")
  }
  paste0(round(100 * x, digits), "%")
}

round_num <- function(x, digits = 3L) {
  ifelse(is.finite(x), round(x, digits), NA_real_)
}

role_palette <- function(role) {
  palette <- c(
    id = "#4dd4ac",
    time = "#63b3ff",
    measure = "#ffd166",
    category = "#f78fb3",
    flag = "#c7f464",
    text = "#b39ddb",
    target = "#ff8a65",
    unknown = "#b9c0c8"
  )
  unname(palette[match(role, names(palette), nomatch = match("unknown", names(palette)))])
}

#' Compare two data DNA profiles
#'
#' @param x A data frame or `data_dna` object.
#' @param y A data frame or `data_dna` object.
#'
#' @return A `dna_comparison` object.
#' @export
#'
#' @examples
#' demo <- dna_example_customers()
#' dna_compare(demo$customers_old, demo$customers_new)
dna_compare <- function(x, y) {
  x_name <- deparse(substitute(x))
  y_name <- deparse(substitute(y))
  x <- as_data_dna(x, name = x_name)
  y <- as_data_dna(y, name = y_name)
  scores <- c(
    schema = schema_similarity(x, y),
    shape = shape_similarity(x, y),
    species = species_similarity(x, y),
    role_structure = role_structure_similarity(x, y),
    distribution = distribution_similarity(x, y),
    missingness = missingness_similarity(x, y),
    category = category_similarity(x, y),
    identity = identity_similarity(x, y)
  )
  weights <- c(
    schema = 0.22,
    shape = 0.08,
    species = 0.10,
    role_structure = 0.15,
    distribution = 0.15,
    missingness = 0.12,
    category = 0.08,
    identity = 0.10
  )
  overall <- weighted_score(scores, weights[names(scores)])
  relation <- if (overall >= 0.97) {
    "same dataset, near-identical version"
  } else if (overall >= 0.85) {
    "same dataset, changed version"
  } else if (overall >= 0.65) {
    "related dataset"
  } else {
    "different dataset"
  }

  structure(
    list(
      x_name = x$name,
      y_name = y$name,
      overall = overall,
      scores = scores,
      likely_relation = relation,
      top_mutations = top_mutation_messages(x, y)
    ),
    class = "dna_comparison"
  )
}

#' @export
print.dna_comparison <- function(x, ...) {
  cat("DataDNA comparison\n")
  cat("  ", x$x_name, " vs ", x$y_name, "\n", sep = "")
  cat("  Overall similarity: ", round(x$overall, 3), "\n", sep = "")
  cat("  Likely relation: ", x$likely_relation, "\n", sep = "")
  for (nm in names(x$scores)) {
    cat("  ", nm, ": ", round(x$scores[[nm]], 3), "\n", sep = "")
  }
  if (length(x$top_mutations)) {
    cat("  Top mutations: ", paste(x$top_mutations, collapse = " | "), "\n", sep = "")
  }
  invisible(x)
}

schema_similarity <- function(x, y) {
  jaccard(paste(x$columns$name, x$columns$type, sep = ":"), paste(y$columns$name, y$columns$type, sep = ":"))
}

shape_similarity <- function(x, y) {
  row_sim <- exp(-abs(log1p(x$n_rows) - log1p(y$n_rows)))
  col_sim <- exp(-abs(log1p(x$n_cols) - log1p(y$n_cols)))
  weighted_score(c(row_sim, col_sim), c(0.6, 0.4))
}

species_similarity <- function(x, y) {
  if (identical(x$species, y$species)) {
    return(1)
  }
  groups <- list(
    entity = c("customer_table", "lookup_table", "survey_data"),
    event = c("transaction_log", "event_stream", "time_series"),
    matrix = c("wide_feature_matrix"),
    unknown = c("unknown_table")
  )
  same_group <- any(vapply(groups, function(g) x$species %in% g && y$species %in% g, logical(1)))
  if (same_group) 0.55 else 0
}

role_structure_similarity <- function(x, y) {
  role_sim <- distribution_overlap(x$columns$role, y$columns$role)
  type_sim <- distribution_overlap(x$columns$type, y$columns$type)
  weighted_score(c(role_sim, type_sim), c(0.7, 0.3))
}

common_columns <- function(x, y) {
  intersect(x$columns$name, y$columns$name)
}

missingness_similarity <- function(x, y) {
  common <- common_columns(x, y)
  if (!length(common)) return(0)
  a <- x$columns[match(common, x$columns$name), ]
  b <- y$columns[match(common, y$columns$name), ]
  clip01(1 - mean(abs(a$missing_rate - b$missing_rate), na.rm = TRUE))
}

identity_similarity <- function(x, y) {
  common <- common_columns(x, y)
  if (!length(common)) return(0)
  a <- x$columns[match(common, x$columns$name), ]
  b <- y$columns[match(common, y$columns$name), ]
  role_sim <- mean(a$role == b$role)
  unique_sim <- clip01(1 - mean(abs(a$unique_rate - b$unique_rate), na.rm = TRUE))
  weighted_score(c(role_sim, unique_sim), c(0.55, 0.45))
}

distribution_similarity <- function(x, y) {
  common <- common_columns(x, y)
  if (!length(common)) return(0)
  a <- x$columns[match(common, x$columns$name), ]
  b <- y$columns[match(common, y$columns$name), ]
  keep <- a$type == "numeric" & b$type == "numeric"
  if (!any(keep)) return(NA_real_)
  qa <- as.matrix(a[keep, c("q25", "q50", "q75"), drop = FALSE])
  qb <- as.matrix(b[keep, c("q25", "q50", "q75"), drop = FALSE])
  scale <- pmax(abs(qa), abs(qb), 1, na.rm = TRUE)
  diff <- abs(qa - qb) / scale
  clip01(1 - mean(diff, na.rm = TRUE))
}

category_similarity <- function(x, y) {
  common <- common_columns(x, y)
  if (!length(common)) return(0)
  a <- x$columns[match(common, x$columns$name), ]
  b <- y$columns[match(common, y$columns$name), ]
  keep <- a$type %in% c("character", "factor", "logical") & b$type %in% c("character", "factor", "logical")
  if (!any(keep)) return(NA_real_)
  sims <- vapply(which(keep), function(i) jaccard(a$top_values[[i]], b$top_values[[i]]), numeric(1))
  mean(sims)
}

distribution_overlap <- function(x, y) {
  tx <- table(x) / length(x)
  ty <- table(y) / length(y)
  keys <- union(names(tx), names(ty))
  px <- as.numeric(tx[keys])
  py <- as.numeric(ty[keys])
  px[is.na(px)] <- 0
  py[is.na(py)] <- 0
  clip01(1 - sum(abs(px - py)) / 2)
}

top_mutation_messages <- function(x, y, n = 3L) {
  changes <- collect_mutations(x, y)
  if (!nrow(changes)) {
    return(character())
  }
  priority <- match(changes$severity, c("high", "medium", "low"), nomatch = 4L)
  changes <- changes[order(priority), , drop = FALSE]
  utils::head(changes$message, n)
}

#' Explain mutations between two data DNA profiles
#'
#' @param x A data frame or `data_dna` object.
#' @param y A data frame or `data_dna` object.
#'
#' @return A `dna_diff` object containing a mutation table.
#' @export
#'
#' @examples
#' demo <- dna_example_customers()
#' dna_diff(demo$customers_old, demo$customers_new)
dna_diff <- function(x, y) {
  x_name <- deparse(substitute(x))
  y_name <- deparse(substitute(y))
  x <- as_data_dna(x, name = x_name)
  y <- as_data_dna(y, name = y_name)
  changes <- collect_mutations(x, y)
  structure(
    list(
      x_name = x$name,
      y_name = y$name,
      changes = changes
    ),
    class = "dna_diff"
  )
}

#' @export
print.dna_diff <- function(x, ...) {
  cat("DataDNA mutation report\n")
  cat("  ", x$x_name, " -> ", x$y_name, "\n", sep = "")
  if (!nrow(x$changes)) {
    cat("  No notable mutations detected.\n")
    return(invisible(x))
  }
  for (i in seq_len(nrow(x$changes))) {
    cat("  [", x$changes$severity[[i]], "] ", x$changes$type[[i]], ": ", x$changes$message[[i]], "\n", sep = "")
  }
  invisible(x)
}

collect_mutations <- function(x, y) {
  old <- x$columns
  new <- y$columns
  old_names <- old$name
  new_names <- new$name
  changes <- list()

  added <- setdiff(new_names, old_names)
  removed <- setdiff(old_names, new_names)
  if (x$n_rows != y$n_rows || x$n_cols != y$n_cols) {
    changes <- append_change(
      changes,
      "shape",
      if (abs(log1p(x$n_rows) - log1p(y$n_rows)) > 0.35 || abs(x$n_cols - y$n_cols) > 3L) "medium" else "low",
      paste0(
        "shape changed from ",
        format(x$n_rows, big.mark = ","),
        "x",
        x$n_cols,
        " to ",
        format(y$n_rows, big.mark = ","),
        "x",
        y$n_cols
      )
    )
  }

  if (!length(intersect(old_names, new_names)) && (length(added) || length(removed))) {
    changes <- append_change(
      changes,
      "schema",
      "high",
      paste0("schema replaced: ", length(removed), " columns removed, ", length(added), " columns added")
    )
  }
  for (nm in utils::head(added, 6L)) changes <- append_change(changes, "schema", "medium", paste0("column added: ", nm))
  for (nm in utils::head(removed, 6L)) changes <- append_change(changes, "schema", "medium", paste0("column removed: ", nm))

  common <- intersect(old_names, new_names)
  for (nm in common) {
    a <- old[match(nm, old$name), ]
    b <- new[match(nm, new$name), ]
    if (!identical(a$type, b$type)) {
      changes <- append_change(changes, "type", "high", paste0(nm, " changed type from ", a$type, " to ", b$type))
    }
    miss_delta <- b$missing_rate - a$missing_rate
    if (is.finite(miss_delta) && abs(miss_delta) >= 0.1) {
      changes <- append_change(changes, "missingness", severity_from_delta(abs(miss_delta)), paste0(nm, " missingness changed by ", format_pct(miss_delta, 1)))
    }
    uniq_delta <- b$unique_rate - a$unique_rate
    if (is.finite(uniq_delta) && abs(uniq_delta) >= 0.15) {
      changes <- append_change(changes, "identity", severity_from_delta(abs(uniq_delta)), paste0(nm, " uniqueness changed by ", format_pct(uniq_delta, 1)))
    }
    if (a$type == "numeric" && b$type == "numeric") {
      shift <- numeric_shift(a, b)
      if (is.finite(shift) && shift >= 0.2) {
        changes <- append_change(changes, "distribution", severity_from_delta(shift), paste0(nm, " numeric distribution shifted"))
      }
    }
    if (a$type %in% c("character", "factor", "logical") && b$type %in% c("character", "factor", "logical")) {
      overlap <- jaccard(a$top_values[[1L]], b$top_values[[1L]])
      if (is.finite(overlap) && overlap < 0.6) {
        changes <- append_change(changes, "category", "medium", paste0(nm, " top categories changed"))
      }
    }
  }

  if (!length(changes)) {
    return(data.frame(type = character(), severity = character(), message = character(), stringsAsFactors = FALSE))
  }
  do.call(rbind, changes)
}

append_change <- function(changes, type, severity, message) {
  c(changes, list(data.frame(type = type, severity = severity, message = message, stringsAsFactors = FALSE)))
}

severity_from_delta <- function(delta) {
  if (delta >= 0.35) "high" else if (delta >= 0.18) "medium" else "low"
}

numeric_shift <- function(a, b) {
  qa <- as.numeric(a[c("q25", "q50", "q75")])
  qb <- as.numeric(b[c("q25", "q50", "q75")])
  scale <- pmax(abs(qa), abs(qb), 1)
  mean(abs(qa - qb) / scale, na.rm = TRUE)
}

#' Render a laboratory-style data DNA card
#'
#' @param x A data frame or `data_dna` object.
#' @param file Optional HTML file path. If supplied, the card is saved there.
#' @param open Logical. Open the saved file in a browser when `file` is supplied.
#'
#' @return An `htmltools` browsable object, invisibly when saved to `file`.
#' @export
#'
#' @examples
#' demo <- dna_example_customers()
#' card <- dna_card(demo$customers_new)
dna_card <- function(x, file = NULL, open = FALSE) {
  dna <- as_data_dna(x)
  card <- build_dna_card(dna)

  if (!is.null(file)) {
    htmltools::save_html(card, file = file)
    if (isTRUE(open)) {
      utils::browseURL(normalizePath(file, winslash = "/", mustWork = FALSE))
    }
    return(invisible(card))
  }

  htmltools::browsable(card)
}

build_dna_card <- function(dna) {
  htmltools::tags$html(
    htmltools::tags$head(
      htmltools::tags$meta(charset = "utf-8"),
      htmltools::tags$title(paste("DataDNA", dna$name)),
      htmltools::tags$style(htmltools::HTML(dna_card_css()))
    ),
    htmltools::tags$body(
      htmltools::tags$main(
        class = "dna-shell",
        htmltools::tags$section(
          class = "dna-card",
          htmltools::tags$div(
            class = "hero",
            htmltools::tags$div(
              htmltools::tags$div(class = "kicker", "DataDNA specimen card"),
              htmltools::tags$h1(dna$name),
              htmltools::tags$p(class = "dna-id", dna$dna_id)
            ),
            htmltools::tags$div(
              class = paste("risk", dna$scores$drift_risk),
              htmltools::tags$span("Drift risk"),
              htmltools::tags$strong(dna$scores$drift_risk)
            )
          ),
          htmltools::tags$div(
            class = "barcode",
            lapply(dna_barcode_values(dna), function(v) htmltools::tags$i(style = paste0("--h:", v, "%;")))
          ),
          fingerprint_stack(dna),
          htmltools::tags$div(
            class = "metrics",
            metric_tile("Species", dna$species),
            metric_tile("Shape", paste0(format(dna$n_rows, big.mark = ","), " x ", dna$n_cols)),
            metric_tile("Identity", paste0(round(dna$scores$identity_strength), "/100")),
            metric_tile("Stability", dna$scores$stability)
          ),
          htmltools::tags$div(
            class = "grid",
            card_panel("Role map", role_map(dna)),
            card_panel("Column signals", column_signal_table(dna)),
            card_panel("Lab notes", lab_notes(dna))
          )
        )
      )
    )
  )
}

metric_tile <- function(label, value) {
  htmltools::tags$div(
    class = "metric",
    htmltools::tags$span(label),
    htmltools::tags$strong(value)
  )
}

card_panel <- function(title, content) {
  htmltools::tags$div(
    class = "panel",
    htmltools::tags$h2(title),
    content
  )
}

dna_barcode_values <- function(dna) {
  text <- paste(dna$fingerprints$schema_hash, dna$fingerprints$profile_hash, dna$dna_id)
  bytes <- as.integer(charToRaw(text))
  vals <- 25 + (bytes %% 70)
  vals[seq_len(min(48L, length(vals)))]
}

role_map <- function(dna) {
  cols <- dna$columns
  htmltools::tags$div(
    htmltools::tags$div(
      class = "role-map",
      lapply(seq_len(nrow(cols)), function(i) {
        htmltools::tags$div(
          class = "gene",
          title = paste(cols$name[[i]], cols$role[[i]], sep = " - "),
          style = paste0("background:", role_palette(cols$role[[i]]), ";")
        )
      })
    ),
    role_legend(cols)
  )
}

fingerprint_stack <- function(dna) {
  htmltools::tags$div(
    class = "fingerprints",
    fingerprint_chip("schema", dna$fingerprints$schema_hash),
    fingerprint_chip("shape", dna$fingerprints$shape_hash),
    fingerprint_chip("profile", dna$fingerprints$profile_hash)
  )
}

fingerprint_chip <- function(label, value) {
  htmltools::tags$div(
    class = "fingerprint-chip",
    htmltools::tags$span(label),
    htmltools::tags$strong(value)
  )
}

role_legend <- function(cols) {
  counts <- sort(table(cols$role), decreasing = TRUE)
  htmltools::tags$div(
    class = "legend",
    lapply(names(counts), function(role) {
      htmltools::tags$span(
        htmltools::tags$i(style = paste0("background:", role_palette(role), ";")),
        paste0(role, " ", counts[[role]])
      )
    })
  )
}

column_signal_table <- function(dna) {
  cols <- dna$columns
  order_idx <- order(cols$missing_rate, decreasing = TRUE, na.last = TRUE)
  show <- cols[head(order_idx, 8L), , drop = FALSE]
  htmltools::tags$table(
    class = "signals-table",
    htmltools::tags$thead(
      htmltools::tags$tr(
        htmltools::tags$th("Column"),
        htmltools::tags$th("Role"),
        htmltools::tags$th("Missing"),
        htmltools::tags$th("Unique")
      )
    ),
    htmltools::tags$tbody(
      lapply(seq_len(nrow(show)), function(i) {
        htmltools::tags$tr(
          htmltools::tags$td(show$name[[i]]),
          htmltools::tags$td(show$role[[i]]),
          htmltools::tags$td(format_pct(show$missing_rate[[i]])),
          htmltools::tags$td(format_pct(show$unique_rate[[i]]))
        )
      })
    )
  )
}

lab_notes <- function(dna) {
  notes <- c(dna$signals, dna$warnings)
  if (!length(notes)) {
    notes <- "No notable mutations detected."
  }
  htmltools::tags$ul(
    class = "notes",
    lapply(head(notes, 8L), htmltools::tags$li)
  )
}

dna_card_css <- function() {
  "
  * { box-sizing: border-box; }
  body {
    margin: 0;
    min-height: 100vh;
    color: #edf7f4;
    background:
      radial-gradient(circle at top left, rgba(77, 212, 172, .16), transparent 32rem),
      linear-gradient(135deg, #111827 0%, #172033 48%, #0f1720 100%);
    font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', sans-serif;
  }
  .dna-shell { width: min(1120px, calc(100vw - 32px)); margin: 32px auto; }
  .dna-card {
    border: 1px solid rgba(166, 244, 214, .28);
    background: rgba(9, 18, 28, .78);
    box-shadow: 0 24px 80px rgba(0, 0, 0, .34);
    border-radius: 8px;
    overflow: hidden;
  }
  .hero {
    display: flex;
    justify-content: space-between;
    gap: 24px;
    padding: 34px;
    border-bottom: 1px solid rgba(166, 244, 214, .16);
  }
  .kicker { color: #7ee7c1; text-transform: uppercase; font-size: 12px; letter-spacing: 0; }
  h1 { margin: 8px 0; font-size: 34px; line-height: 1.1; letter-spacing: 0; }
  .dna-id { margin: 0; color: #b9c7d6; font-family: Consolas, monospace; }
  .risk {
    align-self: flex-start;
    min-width: 130px;
    padding: 14px 16px;
    border-radius: 8px;
    background: rgba(255, 255, 255, .06);
    border: 1px solid rgba(255, 255, 255, .12);
  }
  .risk span { display: block; color: #b9c7d6; font-size: 12px; }
  .risk strong { display: block; margin-top: 4px; text-transform: uppercase; }
  .risk.low strong { color: #7ee7c1; }
  .risk.medium strong { color: #ffd166; }
  .risk.high strong { color: #ff8a65; }
  .barcode {
    display: flex;
    align-items: end;
    gap: 4px;
    height: 98px;
    padding: 20px 34px;
    background: rgba(255, 255, 255, .035);
  }
  .barcode i {
    display: block;
    width: 10px;
    height: var(--h);
    border-radius: 3px 3px 0 0;
    background: linear-gradient(180deg, #7ee7c1, #63b3ff);
  }
  .fingerprints {
    display: grid;
    grid-template-columns: repeat(3, 1fr);
    gap: 1px;
    background: rgba(166, 244, 214, .14);
    border-top: 1px solid rgba(166, 244, 214, .12);
  }
  .fingerprint-chip {
    padding: 12px 24px;
    background: rgba(12, 25, 39, .94);
    font-family: Consolas, monospace;
  }
  .fingerprint-chip span {
    display: block;
    color: #9eb1c3;
    font-size: 11px;
    text-transform: uppercase;
  }
  .fingerprint-chip strong {
    display: block;
    margin-top: 5px;
    color: #d9fff2;
    font-size: 14px;
    letter-spacing: 0;
  }
  .metrics {
    display: grid;
    grid-template-columns: repeat(4, 1fr);
    gap: 1px;
    background: rgba(166, 244, 214, .14);
  }
  .metric { padding: 20px 24px; background: rgba(12, 25, 39, .96); }
  .metric span { display: block; color: #9eb1c3; font-size: 12px; }
  .metric strong { display: block; margin-top: 6px; font-size: 19px; }
  .grid { display: grid; grid-template-columns: 1fr 1.35fr 1fr; gap: 18px; padding: 24px; }
  .panel {
    min-width: 0;
    padding: 18px;
    border: 1px solid rgba(255, 255, 255, .1);
    border-radius: 8px;
    background: rgba(255, 255, 255, .045);
  }
  h2 { margin: 0 0 14px; font-size: 15px; letter-spacing: 0; }
  .role-map { display: grid; grid-template-columns: repeat(auto-fit, minmax(18px, 1fr)); gap: 7px; }
  .gene { height: 40px; border-radius: 5px; box-shadow: inset 0 -8px 14px rgba(0, 0, 0, .18); }
  .legend {
    display: flex;
    flex-wrap: wrap;
    gap: 8px;
    margin-top: 14px;
    color: #cbd7e2;
    font-size: 12px;
  }
  .legend span {
    display: inline-flex;
    align-items: center;
    gap: 6px;
    padding: 5px 8px;
    border-radius: 999px;
    background: rgba(255, 255, 255, .06);
  }
  .legend i {
    display: inline-block;
    width: 9px;
    height: 9px;
    border-radius: 50%;
  }
  .signals-table { width: 100%; border-collapse: collapse; font-size: 13px; }
  .signals-table th { color: #9eb1c3; font-weight: 600; text-align: left; }
  .signals-table th, .signals-table td { padding: 8px 6px; border-bottom: 1px solid rgba(255, 255, 255, .08); }
  .notes { margin: 0; padding-left: 18px; color: #dbe8e4; }
  .notes li { margin: 0 0 9px; }
  @media (max-width: 820px) {
    .hero, .grid { grid-template-columns: 1fr; display: grid; }
    .fingerprints { grid-template-columns: 1fr; }
    .metrics { grid-template-columns: repeat(2, 1fr); }
    h1 { font-size: 27px; }
  }
  "
}

#' Old customer table for DataDNA examples
#'
#' A small synthetic customer table used to demonstrate data DNA profiling.
#'
#' @format A data frame with 180 rows and 8 columns.
#' @source Synthetic data generated for package examples.
"customers_old"

#' New customer table for DataDNA examples
#'
#' A modified version of `customers_old` with distribution, category,
#' missingness, and schema changes.
#'
#' @format A data frame with 180 rows and 9 columns.
#' @source Synthetic data generated for package examples.
"customers_new"

#' Draw a paper-style lineage figure
#'
#' Creates a print-friendly, paper-style lineage figure for a `dna_match` object
#' using base R grid graphics. The figure can be drawn on the current graphics
#' device or saved directly to PNG or PDF.
#'
#' @param match A `dna_match` object.
#' @param file Optional output path. Supported extensions are `.png` and `.pdf`.
#' @param width Plot width in inches.
#' @param height Plot height in inches.
#' @param dpi Resolution used for PNG output.
#'
#' @return The input `dna_match` object, invisibly.
#' @export
#'
#' @examples
#' demo <- dna_example_customers()
#' lib <- list(old = data_dna(demo$customers_old), new = data_dna(demo$customers_new))
#' match <- dna_match(demo$customers_new, lib)
#' dna_match_plot(match)
dna_match_plot <- function(match, file = NULL, width = 11, height = 7, dpi = 144) {
  if (!inherits(match, "dna_match")) {
    stop("`match` must be a dna_match object.", call. = FALSE)
  }

  if (!is.null(file)) {
    ext <- tolower(tools::file_ext(file))
    if (!ext %in% c("png", "pdf")) {
      stop("`file` must end in .png or .pdf.", call. = FALSE)
    }
    if (ext == "png") {
      grDevices::png(file, width = width, height = height, units = "in", res = dpi, bg = "#ffffff")
    } else {
      grDevices::pdf(file, width = width, height = height, bg = "#ffffff", useDingbats = FALSE)
    }
    on.exit(grDevices::dev.off(), add = TRUE)
  }

  draw_match_report(match)
  invisible(match)
}

draw_match_report <- function(match) {
  grid::grid.newpage()
  layout <- report_layout()
  palette <- report_palette()
  best <- match$matches[1L, , drop = FALSE]

  grid::grid.rect(gp = grid::gpar(fill = palette$background, col = NA))
  draw_compact_header(match, best, layout, palette)
  draw_lineage_figure(match, layout, palette)
  draw_compact_table(match, layout, palette)
  draw_numbered_notes(match, layout, palette)
}

report_layout <- function() {
  list(
    margin_left = 0.055,
    margin_right = 0.055,
    header_y = 0.91,
    meta_y = 0.855,
    graph = c(x = 0.055, y = 0.25, w = 0.60, h = 0.50),
    table = c(x = 0.69, y = 0.25, w = 0.255, h = 0.50),
    notes = c(x = 0.055, y = 0.06, w = 0.89, h = 0.13)
  )
}

draw_compact_header <- function(match, best, layout, palette) {
  grid::grid.text(
    "Figure 1. DataDNA lineage match",
    x = grid::unit(layout$margin_left, "npc"),
    y = grid::unit(layout$header_y, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$text, fontsize = 16, fontface = "bold", fontfamily = report_font())
  )

  grid::grid.text(
    paste0(
      "Query: ",
      match$query$name,
      "   |   ",
      match$query$species,
      "   |   ",
      format(match$query$n_rows, big.mark = ","),
      " x ",
      match$query$n_cols
    ),
    x = grid::unit(layout$margin_left, "npc"),
    y = grid::unit(layout$meta_y, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$muted, fontsize = 9.5, fontfamily = report_font())
  )
  grid::grid.text(
    paste0("DNA ID  ", match$query$dna_id),
    x = grid::unit(layout$margin_left, "npc"),
    y = grid::unit(layout$meta_y - 0.032, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$faint, fontsize = 8.6, fontfamily = report_mono_font())
  )

  draw_panel(0.70, 0.842, 0.245, 0.095, palette)
  grid::grid.text(
    "Best match",
    x = grid::unit(0.718, "npc"),
    y = grid::unit(0.908, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$muted, fontsize = 8.5, fontface = "bold", fontfamily = report_font())
  )
  grid::grid.text(
    plot_short_label(best$name, 20L),
    x = grid::unit(0.718, "npc"),
    y = grid::unit(0.882, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$text, fontsize = 12, fontface = "bold", fontfamily = report_font())
  )
  grid::grid.text(
    format_pct(best$overall),
    x = grid::unit(0.925, "npc"),
    y = grid::unit(0.884, "npc"),
    just = c("right", "center"),
    gp = grid::gpar(col = palette$accent, fontsize = 17, fontface = "bold", fontfamily = report_font())
  )
  grid::grid.text(
    plot_short_label(best$relation, 32L),
    x = grid::unit(0.718, "npc"),
    y = grid::unit(0.858, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$muted, fontsize = 8.5, fontfamily = report_font())
  )
}

draw_lineage_figure <- function(match, layout, palette) {
  g <- layout$graph
  draw_panel(g[["x"]], g[["y"]], g[["w"]], g[["h"]], palette)

  grid::grid.text(
    "A. Similarity-weighted lineage",
    x = grid::unit(g[["x"]] + 0.025, "npc"),
    y = grid::unit(g[["y"]] + g[["h"]] - 0.04, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$text, fontsize = 11.5, fontface = "bold", fontfamily = report_font())
  )

  matches <- match$matches
  n <- nrow(matches)
  qx <- g[["x"]] + 0.055
  qy <- g[["y"]] + g[["h"]] * 0.48
  qw <- 0.19
  qh <- 0.105
  cx <- g[["x"]] + 0.355
  cw <- 0.205
  ch <- 0.076
  cy <- if (n == 1L) qy else seq(g[["y"]] + g[["h"]] - 0.115, g[["y"]] + 0.10, length.out = n)

  for (i in seq_len(n)) {
    score <- matches$overall[[i]]
    col <- relation_color_static(score)
    grid::grid.xspline(
      x = grid::unit(c(qx + qw, qx + qw + 0.07, cx - 0.07, cx), "npc"),
      y = grid::unit(c(qy, qy, cy[[i]], cy[[i]]), "npc"),
      shape = 1,
      gp = grid::gpar(col = col, lwd = 0.8 + 2.2 * clip01(score), alpha = 0.82)
    )
  }

  draw_lineage_node(qx, qy, qw, qh, match$query$name, match$query$dna_id, match$query$species, palette$accent, palette, TRUE)
  for (i in seq_len(n)) {
    score <- matches$overall[[i]]
    draw_lineage_node(
      cx,
      cy[[i]],
      cw,
      ch,
      matches$name[[i]],
      plot_relation_label(matches$relation[[i]]),
      format_pct(score),
      relation_color_static(score),
      palette,
      FALSE
    )
  }
}

draw_compact_table <- function(match, layout, palette) {
  t <- layout$table
  draw_panel(t[["x"]], t[["y"]], t[["w"]], t[["h"]], palette)

  grid::grid.text(
    "B. Ranked candidates",
    x = grid::unit(t[["x"]] + 0.018, "npc"),
    y = grid::unit(t[["y"]] + t[["h"]] - 0.04, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$text, fontsize = 11.5, fontface = "bold", fontfamily = report_font())
  )

  header_y <- t[["y"]] + t[["h"]] - 0.085
  col_x <- c(t[["x"]] + 0.018, t[["x"]] + 0.07, t[["x"]] + t[["w"]] - 0.018)
  headers <- c("#", "candidate", "similarity")
  justs <- list(c("left", "center"), c("left", "center"), c("right", "center"))
  for (i in seq_along(headers)) {
    grid::grid.text(
      headers[[i]],
      x = grid::unit(col_x[[i]], "npc"),
      y = grid::unit(header_y, "npc"),
      just = justs[[i]],
      gp = grid::gpar(col = palette$muted, fontsize = 8, fontface = "bold", fontfamily = report_font())
    )
  }
  grid::grid.lines(
    x = grid::unit(c(t[["x"]] + 0.018, t[["x"]] + t[["w"]] - 0.018), "npc"),
    y = grid::unit(c(header_y - 0.025, header_y - 0.025), "npc"),
    gp = grid::gpar(col = palette$border, lwd = 1)
  )

  matches <- match$matches
  row_y <- header_y - 0.055
  row_gap <- min(0.084, (t[["h"]] - 0.15) / max(nrow(matches), 1L))
  for (i in seq_len(nrow(matches))) {
    score <- matches$overall[[i]]
    grid::grid.text(sprintf("%02d", i), x = grid::unit(col_x[[1L]], "npc"), y = grid::unit(row_y, "npc"), just = justs[[1L]], gp = grid::gpar(col = palette$faint, fontsize = 8.5, fontfamily = report_mono_font()))
    grid::grid.text(plot_short_label(matches$name[[i]], 19L), x = grid::unit(col_x[[2L]], "npc"), y = grid::unit(row_y + 0.01, "npc"), just = justs[[2L]], gp = grid::gpar(col = palette$text, fontsize = 9.2, fontface = "bold", fontfamily = report_font()))
    grid::grid.text(format_pct(score), x = grid::unit(col_x[[3L]], "npc"), y = grid::unit(row_y + 0.01, "npc"), just = justs[[3L]], gp = grid::gpar(col = relation_color_static(score), fontsize = 10, fontface = "bold", fontfamily = report_font()))
    grid::grid.text(plot_relation_label(matches$relation[[i]]), x = grid::unit(col_x[[2L]], "npc"), y = grid::unit(row_y - 0.015, "npc"), just = justs[[2L]], gp = grid::gpar(col = palette$muted, fontsize = 8.1, fontfamily = report_font()))

    meter_x <- t[["x"]] + 0.07
    meter_w <- t[["w"]] - 0.088
    grid::grid.rect(x = grid::unit(meter_x, "npc"), y = grid::unit(row_y - 0.04, "npc"), width = grid::unit(meter_w, "npc"), height = grid::unit(0.003, "npc"), just = c("left", "center"), gp = grid::gpar(fill = palette$panel, col = NA))
    grid::grid.rect(x = grid::unit(meter_x, "npc"), y = grid::unit(row_y - 0.04, "npc"), width = grid::unit(meter_w * clip01(score), "npc"), height = grid::unit(0.003, "npc"), just = c("left", "center"), gp = grid::gpar(fill = relation_color_static(score), col = NA))

    row_y <- row_y - row_gap
  }
}

draw_numbered_notes <- function(match, layout, palette) {
  n <- layout$notes
  best <- match$matches[1L, , drop = FALSE]
  notes <- best$top_mutations[[1L]]
  if (!length(notes)) {
    notes <- "No notable mutations detected for the closest relative."
  }
  notes <- utils::head(notes, 3L)

  draw_panel(n[["x"]], n[["y"]], n[["w"]], n[["h"]], palette)
  grid::grid.text(
    "C. Mutation notes",
    x = grid::unit(n[["x"]] + 0.025, "npc"),
    y = grid::unit(n[["y"]] + n[["h"]] - 0.035, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$text, fontsize = 11.5, fontface = "bold", fontfamily = report_font())
  )
  grid::grid.text(
    paste0(best$name, " is the nearest non-self relative (", format_pct(best$overall), " similarity)."),
    x = grid::unit(n[["x"]] + 0.025, "npc"),
    y = grid::unit(n[["y"]] + n[["h"]] - 0.07, "npc"),
    just = c("left", "center"),
    gp = grid::gpar(col = palette$muted, fontsize = 9, fontfamily = report_font())
  )

  y <- n[["y"]] + 0.035
  for (i in seq_along(notes)) {
    x <- n[["x"]] + 0.025 + (i - 1L) * (n[["w"]] - 0.05) / length(notes)
    grid::grid.text(
      paste0(i, ". ", plot_short_label(notes[[i]], 34L)),
      x = grid::unit(x, "npc"),
      y = grid::unit(y, "npc"),
      just = c("left", "center"),
      gp = grid::gpar(col = palette$text, fontsize = 8.8, fontfamily = report_font())
    )
  }
}

draw_lineage_node <- function(x, y, w, h, title, subtitle, meta, accent, palette, query = FALSE) {
  fill <- if (query) palette$panel_dark else palette$background
  text_col <- if (query) palette$text_reverse else palette$text
  sub_col <- if (query) "#c9d3dd" else palette$muted
  draw_round_rect(x, y - h / 2, w, h, fill = fill, col = accent)
  grid::grid.rect(x = grid::unit(x, "npc"), y = grid::unit(y, "npc"), width = grid::unit(0.0035, "npc"), height = grid::unit(h - 0.03, "npc"), just = c("left", "center"), gp = grid::gpar(fill = accent, col = NA))
  grid::grid.text(plot_short_label(title, 22L), x = grid::unit(x + 0.016, "npc"), y = grid::unit(y + h * 0.18, "npc"), just = c("left", "center"), gp = grid::gpar(col = text_col, fontsize = if (query) 10.5 else 9.2, fontface = "bold", fontfamily = report_font()))
  grid::grid.text(plot_short_label(subtitle, if (query) 27L else 30L), x = grid::unit(x + 0.016, "npc"), y = grid::unit(y - h * 0.03, "npc"), just = c("left", "center"), gp = grid::gpar(col = sub_col, fontsize = 7.6, fontfamily = if (query) report_mono_font() else report_font()))
  grid::grid.text(plot_short_label(meta, 26L), x = grid::unit(x + 0.016, "npc"), y = grid::unit(y - h * 0.25, "npc"), just = c("left", "center"), gp = grid::gpar(col = sub_col, fontsize = 7.6, fontfamily = report_font()))
}

draw_panel <- function(x, y, w, h, palette) {
  draw_round_rect(x, y, w, h, fill = palette$background, col = palette$border)
}

draw_round_rect <- function(x, y, w, h, fill, col) {
  grid::grid.roundrect(
    x = grid::unit(x, "npc"),
    y = grid::unit(y, "npc"),
    width = grid::unit(w, "npc"),
    height = grid::unit(h, "npc"),
    just = c("left", "bottom"),
    r = grid::unit(0.006, "npc"),
    gp = grid::gpar(fill = fill, col = col, lwd = 0.8)
  )
}

relation_color_static <- function(score) {
  if (score >= 0.97) {
    "#2f9e7e"
  } else if (score >= 0.85) {
    "#2b7bb9"
  } else if (score >= 0.65) {
    "#b08a1d"
  } else {
    "#8a949e"
  }
}

plot_short_label <- function(x, width = 26L) {
  x <- as.character(x)
  ifelse(nchar(x, type = "width") > width, paste0(substr(x, 1L, width - 1L), "..."), x)
}

plot_relation_label <- function(x) {
  x <- as.character(x)
  out <- x
  out[grepl("near-identical", x, fixed = TRUE)] <- "near-identical version"
  out[grepl("changed version", x, fixed = TRUE)] <- "changed version"
  out[grepl("related dataset", x, fixed = TRUE)] <- "related dataset"
  out[grepl("different dataset", x, fixed = TRUE)] <- "different dataset"
  out
}

report_palette <- function() {
  list(
    text = "#1f2933",
    text_reverse = "#f8fafc",
    muted = "#6b7280",
    faint = "#9aa8b5",
    accent = "#2b7bb9",
    border = "#d7dde3",
    background = "#ffffff",
    panel = "#f8fafc",
    panel_dark = "#111827"
  )
}

report_font <- function() {
  "sans"
}

report_mono_font <- function() {
  "mono"
}

\name{customers_new}
\alias{customers_new}
\docType{data}
\title{New customer table for DataDNA examples}
\format{
A data frame with 180 rows and 9 columns.
}
\source{
Synthetic data generated for package examples.
}
\description{
A modified version of \code{customers_old} with distribution, category,
missingness, and schema changes.
}
\keyword{datasets}

\name{dna_compare}
\alias{dna_compare}
\title{Compare two data DNA profiles}
\usage{
dna_compare(x, y)
}
\arguments{
\item{x}{A data frame or \code{data_dna} object.}
\item{y}{A data frame or \code{data_dna} object.}
}
\value{
A \code{dna_comparison} object.
}
\description{
Compare two data DNA profiles.
}
\examples{
demo <- dna_example_customers()
dna_compare(demo$customers_old, demo$customers_new)
}

\name{dna_card}
\alias{dna_card}
\title{Render a laboratory-style data DNA card}
\usage{
dna_card(x, file = NULL, open = FALSE)
}
\arguments{
\item{x}{A data frame or \code{data_dna} object.}
\item{file}{Optional HTML file path. If supplied, the card is saved there.}
\item{open}{Logical. Open the saved file in a browser when \code{file} is supplied.}
}
\value{
An \code{htmltools} browsable object, invisibly when saved to \code{file}.
}
\description{
Render a laboratory-style data DNA card.
}
\examples{
demo <- dna_example_customers()
card <- dna_card(demo$customers_new)
}

\name{dna_example_customers}
\alias{dna_example_customers}
\title{Example customer tables}
\usage{
dna_example_customers()
}
\value{
A list with \code{customers_old} and \code{customers_new} data frames.
}
\description{
Creates two small customer data frames designed to demonstrate DataDNA cards,
comparison, and mutation reports.
}
\examples{
demo <- dna_example_customers()
str(demo$customers_old)
}

\name{dna_species}
\alias{dna_species}
\title{Guess the species of a data frame}
\usage{
dna_species(df)
}
\arguments{
\item{df}{A data frame.}
}
\value{
A character label such as \code{customer_table}, \code{event_stream}, or
\code{wide_feature_matrix}.
}
\description{
Guess the species of a data frame.
}
\examples{
dna_species(dna_example_customers()$customers_new)
}

\name{customers_old}
\alias{customers_old}
\docType{data}
\title{Old customer table for DataDNA examples}
\format{
A data frame with 180 rows and 8 columns.
}
\source{
Synthetic data generated for package examples.
}
\description{
A small synthetic customer table used to demonstrate data DNA profiling.
}
\keyword{datasets}

\name{data_dna}
\alias{data_dna}
\title{Create a data DNA profile}
\usage{
data_dna(df, name = NULL, sample_size = 10000L)
}
\arguments{
\item{df}{A data frame.}
\item{name}{Optional data set name shown on cards and print output.}
\item{sample_size}{Maximum number of rows used for profiling.}
}
\value{
A \code{data_dna} object.
}
\description{
Profiles an R data frame into a compact identity object that records schema,
shape, missingness, distributions, categories, uniqueness, time signals, and
stable fingerprints.
}
\examples{
demo <- dna_example_customers()
dna <- data_dna(demo$customers_new, name = "customers_new")
dna
}

\name{dna_match_plot}
\alias{dna_match_plot}
\title{Draw a paper-style lineage figure}
\usage{
dna_match_plot(match, file = NULL, width = 11, height = 7, dpi = 144)
}
\arguments{
\item{match}{A \code{dna_match} object.}
\item{file}{Optional output path. Supported extensions are \code{.png} and \code{.pdf}.}
\item{width}{Plot width in inches.}
\item{height}{Plot height in inches.}
\item{dpi}{Resolution used for PNG output.}
}
\value{
The input \code{dna_match} object, invisibly.
}
\description{
Creates a print-friendly, paper-style lineage figure for a \code{dna_match}
object using base R grid graphics. The figure can be drawn on the current
graphics device or saved directly to PNG or PDF.
}
\examples{
demo <- dna_example_customers()
lib <- list(old = data_dna(demo$customers_old), new = data_dna(demo$customers_new))
match <- dna_match(demo$customers_new, lib)
dna_match_plot(match)
}

\name{dna_match_card}
\alias{dna_match_card}
\title{Render a DataDNA lineage match card}
\usage{
dna_match_card(match, file = NULL, open = FALSE)
}
\arguments{
\item{match}{A \code{dna_match} object.}
\item{file}{Optional HTML file path. If supplied, the card is saved there.}
\item{open}{Logical. Open the saved file in a browser when \code{file} is supplied.}
}
\value{
An \code{htmltools} browsable object, invisibly when saved to \code{file}.
}
\description{
Creates a static HTML/SVG lineage diagram for a \code{dna_match} object.
}
\examples{
demo <- dna_example_customers()
lib <- list(old = data_dna(demo$customers_old), new = data_dna(demo$customers_new))
match <- dna_match(demo$customers_new, lib)
dna_match_card(match)
}

\name{dna_match}
\alias{dna_match}
\title{Match a data set against a DNA library}
\usage{
dna_match(x, library, top_n = 5L, sample_size = 10000L)
}
\arguments{
\item{x}{A data frame or \code{data_dna} object to match.}
\item{library}{A list of data frames or \code{data_dna} objects.}
\item{top_n}{Maximum number of matches to return.}
\item{sample_size}{Maximum number of rows used when profiling raw data frames.}
}
\value{
A \code{dna_match} object.
}
\description{
Finds the closest relatives of a query data set by comparing its data DNA
against a named library of data frames or \code{data_dna} objects.
}
\examples{
demo <- dna_example_customers()
lib <- list(old = data_dna(demo$customers_old), new = data_dna(demo$customers_new))
dna_match(demo$customers_new, lib)
}
